from abc import ABC, abstractmethod


class FeatureFlagRouter(ABC):
    """
    The base feature flag router.
    All feature flag routers should extend this class.
    """

    def _notify_change(
        self, name: str, new_value: bool, old_value: bool | None
    ) -> None:
        """
        Override to provide a method to be used to notify when a feature is enabled or disabled.
        Implementation of when and whether this is called is the responsibility of subclasses.
        This is never called by default; the base class implementation is a no-op.

        :param str name: The name of the feature flag.
        :param bool new_value: The value that the flag is changing to.
        :param bool | None old_value: The value that the flag is changing from.
        """

    @abstractmethod
    def set_feature_is_enabled(self, name: str, is_enabled: bool) -> None:
        """
        Enable or disable a feature flag.

        :param str name: The name of the feature flag.
        :param bool is_enabled: If `True`, the feature is enabled. If `False`, the feature is disabled.
        """

    @abstractmethod
    def feature_is_enabled(self, name: str, default: bool = False) -> bool:
        """
        Determine whether a feature flag is enabled or disabled.

        :param str name: The name of the feature flag.
        :param bool default: A default value to return for cases where a feature flag may not exist. Defaults to False.
        :return bool: If `True`, the feature is enabled. If `False`, the feature is disabled.
        """



from typing import Sequence

from injector import Injector
from Ligare.platform.dependency_injection import UserLoaderModule
from Ligare.platform.identity import Role as DbRole
from Ligare.platform.identity import User as DbUser
from Ligare.platform.identity.user_loader import (
    Role,
    TRole,
    UserId,
    UserLoader,
    UserMixin,
)
from typing_extensions import override


def test__UserLoaderModule__uses_application_loader_mixin():
    class UM(UserMixin[TRole]):
        @override
        def __init__(
            self, user_id: UserId, roles: Sequence[TRole] | None = None
        ) -> None:
            self._user_id = user_id
            self._roles = roles
            super().__init__(user_id, roles)

    class UserTable(DbUser):
        pass

    class RoleTable(DbRole):
        pass

    user_loader_module = UserLoaderModule(
        loader=UM, roles=Role, user_table=UserTable, role_table=RoleTable
    )

    injector = Injector([user_loader_module])

    user_loader = injector.get(UserLoader[UserMixin[Role]])

    assert user_loader._loader is UM  # pyright: ignore[reportPrivateUsage]



from logging import Logger

from typing_extensions import override

from .feature_flag_router import FeatureFlagRouter


class CachingFeatureFlagRouter(FeatureFlagRouter):
    def __init__(self, logger: Logger) -> None:
        self._logger: Logger = logger
        self._feature_flags: dict[str, bool] = {}
        super().__init__()

    @override
    def _notify_change(
        self, name: str, new_value: bool, old_value: bool | None
    ) -> None:
        if name in self._feature_flags:
            if new_value == old_value:
                self._logger.warning(
                    f"Tried to change feature flag value for '{name}' to the same value. It is already {'enabled' if new_value else 'disabled'}."
                )
            else:
                self._logger.warning(
                    f"Changing feature flag value for '{name}' from `{old_value}` to `{new_value}`."
                )
        else:
            self._logger.warning(f"Setting new feature flag '{name}' to `{new_value}`.")

    def _validate_name(self, name: str):
        if type(name) != str:
            raise TypeError("`name` must be a string.")

        if not name:
            raise ValueError("`name` parameter is required and cannot be empty.")

    @override
    def set_feature_is_enabled(self, name: str, is_enabled: bool) -> None:
        """
         Enables or disables a feature flag in the in-memory dictionary of feature flags.

         Subclasses should call this method to validate parameters and cache values.

        :param str name: The feature flag to check.
        :param bool is_enabled: Whether the feature flag is to be enabled or disabled.
        """
        self._validate_name(name)

        if type(is_enabled) != bool:
            raise TypeError("`is_enabled` must be a boolean.")

        self._notify_change(name, is_enabled, self._feature_flags.get(name))

        self._feature_flags[name] = is_enabled

        return super().set_feature_is_enabled(name, is_enabled)

    @override
    def feature_is_enabled(self, name: str, default: bool = False) -> bool:
        """
        Determine whether a feature flag is enabled or disabled.

        Subclasses should call this method to validate parameters and use cached values.

        :param str name: The feature flag to check.
        :param bool default: If the feature flag is not in the in-memory dictionary of flags,
            this is the default value to return. The default parameter value
            when not specified is `False`.
        :return bool: If `True`, the feature is enabled. If `False`, the feature is disabled.
        """
        self._validate_name(name)

        if type(default) != bool:
            raise TypeError("`default` must be a boolean.")

        return self._feature_flags.get(name, default)

    def feature_is_cached(self, name: str):
        self._validate_name(name)

        return name in self._feature_flags



from logging import Logger

from flask import Config as Config
from injector import Binder, Module, provider
from Ligare.database.dependency_injection import ScopedSessionModule
from Ligare.database.types import MetaBase
from Ligare.platform.identity import Role as DbRole
from Ligare.platform.identity import User as DbUser
from Ligare.platform.identity.user_loader import Role, UserLoader, UserMixin
from Ligare.programming.patterns.dependency_injection import LoggerModule
from sqlalchemy.orm.scoping import ScopedSession
from typing_extensions import override


class UserLoaderModule(Module):
    _bases: list[MetaBase | type[MetaBase]] | None = None

    def __init__(
        self,
        loader: type[UserMixin[Role]],
        roles: type[Role],
        user_table: type[DbUser],
        role_table: type[DbRole],
        bases: list[MetaBase | type[MetaBase]] | None = None,
    ) -> None:
        self._loader = loader
        self._roles = roles
        self._bases = bases
        self._user_table = user_table
        self._role_table = role_table
        super().__init__()

    @override
    def configure(self, binder: Binder) -> None:
        binder.install(ScopedSessionModule(self._bases))
        binder.install(LoggerModule)

    @provider
    def _provide_user_mixin_loader(  # pyright: ignore[reportUnknownParameterType]
        self, user_loader: UserLoader[UserMixin[Role]]
    ) -> UserLoader[UserMixin]:  # pyright: ignore[reportMissingTypeArgument]
        """
        The weirdness with this method's return type supports usage of the `UserLoader`
        class, which uses `TUserMixin`, which is a TypeVar bound to an
        invariant type (`UserMixin[Role]`).
        That, in turn, supports `LoginManager` using `UserLoader[UserMixin]` as a dependency.
        This is all done because we cannot use a TypeVar for a type registered with Injector,
        but we can use concrete types for generics.
        """
        return user_loader

    @provider
    def _provide_user_mixin_role_loader(
        self, scoped_session: ScopedSession, log: Logger
    ) -> UserLoader[UserMixin[Role]]:
        """
        `UserLoader[UserMixin[Role]]` is the concrete type that is returned to
        dependents of `UserLoader[UserMixin]`.
        """
        return UserLoader(
            loader=self._loader,
            roles=self._roles,
            user_table=self._user_table,
            role_table=self._role_table,
            scoped_session=scoped_session,
            log=log,
        )



from abc import ABC
from logging import Logger
from typing import Type, cast, overload

from injector import inject
from sqlalchemy import Boolean, Column, Unicode
from sqlalchemy.exc import NoResultFound
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm.session import Session
from typing_extensions import override

from .caching_feature_flag_router import CachingFeatureFlagRouter


class FeatureFlag(ABC):
    def __init__(  # pyright: ignore[reportMissingSuperCall]
        self,
        /,
        name: str,
        description: str,
        enabled: bool | None = False,
    ) -> None:
        raise NotImplementedError(
            f"`{FeatureFlag.__class__.__name__}` should only be used for type checking."
        )

    __tablename__: str
    name: Column[Unicode] | str
    description: Column[Unicode] | str
    enabled: Column[Boolean] | bool


class FeatureFlagTable:
    def __new__(cls, base: Type[DeclarativeMeta]) -> type[FeatureFlag]:
        class _FeatureFlag(base):
            """
            A feature flag.
            """

            __tablename__ = "feature_flag"

            name: Column[Unicode] | str = Column(
                "name", Unicode, primary_key=True, nullable=False
            )
            description: Column[Unicode] | str = Column(
                "description", Unicode, nullable=False
            )
            enabled: Column[Boolean] | bool = Column(
                "enabled", Boolean, nullable=True, default=False
            )

            @override
            def __repr__(self) -> str:
                return "<FeatureFlag %s>" % (self.name)

        return cast(type[FeatureFlag], _FeatureFlag)


class DBFeatureFlagRouter(CachingFeatureFlagRouter):
    _feature_flag: type[FeatureFlag]
    _session: Session

    @inject
    def __init__(
        self, feature_flag: type[FeatureFlag], session: Session, logger: Logger
    ) -> None:
        self._feature_flag = feature_flag
        self._session = session
        super().__init__(logger)

    @override
    def set_feature_is_enabled(self, name: str, is_enabled: bool):
        """
        Enable or disable a feature flag in the database.

        This method caches the value of `is_enabled` for the specified feature flag
        unless saving to the database fails.

        name: The feature flag to check.

        is_enabled: Whether the feature flag is to be enabled or disabled.
        """

        if type(name) != str:
            raise TypeError("`name` must be a string.")

        if not name:
            raise ValueError("`name` parameter is required and cannot be empty.")

        feature_flag: FeatureFlag
        try:
            feature_flag = (
                self._session.query(self._feature_flag)
                .filter(cast(Column[Unicode], self._feature_flag.name) == name)
                .one()
            )
        except NoResultFound as e:
            raise LookupError(
                f"The feature flag `{name}` does not exist. It must be created before being accessed."
            ) from e

        feature_flag.enabled = is_enabled
        self._session.commit()
        super().set_feature_is_enabled(name, is_enabled)

    @overload
    def feature_is_enabled(self, name: str, default: bool = False) -> bool: ...
    @overload
    def feature_is_enabled(
        self, name: str, default: bool, check_cache: bool = True
    ) -> bool: ...
    @override
    def feature_is_enabled(
        self, name: str, default: bool = False, check_cache: bool = True
    ) -> bool:
        """
        Determine whether a feature flag is enabled or disabled.
        This method returns False if the feature flag does not exist in the database.

        This method caches the value pulled from the database
        for the specified feature flag. It is only cached if the value is
        pulled from the database. If the flag does not exist, no value is cached.

        :param str name: The feature flag to check.
        :param bool default: The default value to return when a flag does not exist.
        :param bool check_cache: Whether to use the cached value if it is cached. The default is `True`.
            If the cache is not checked, the new value pulled from the database will be cached.
        """
        if check_cache and super().feature_is_cached(name):
            return super().feature_is_enabled(name, default)

        feature_flag = (
            self._session.query(self._feature_flag)
            .filter(cast(Column[Unicode], self._feature_flag.name) == name)
            .one_or_none()
        )

        if feature_flag is None:
            self._logger.warning(
                f'Feature flag {name} not found in database. Returning "{default}" by default.'
            )
            return default

        is_enabled = cast(bool, feature_flag.enabled)

        super().set_feature_is_enabled(name, is_enabled)

        return is_enabled



from .caching_feature_flag_router import CachingFeatureFlagRouter
from .db_feature_flag_router import DBFeatureFlagRouter
from .feature_flag_router import FeatureFlagRouter

__all__ = ("FeatureFlagRouter", "CachingFeatureFlagRouter", "DBFeatureFlagRouter")



import logging
from typing import Any, Tuple

import pytest
from Ligare.database.config import DatabaseConfig
from Ligare.database.dependency_injection import ScopedSessionModule
from Ligare.platform.feature_flag.db_feature_flag_router import (
    DBFeatureFlagRouter,
    FeatureFlagTable,
)
from Ligare.programming.dependency_injection import ConfigModule
from pytest_mock import MockerFixture
from sqlalchemy.ext.declarative import DeclarativeMeta, declarative_base
from sqlalchemy.orm.session import Session

_FEATURE_FLAG_TEST_NAME = "foo_feature"
_FEATURE_FLAG_TEST_DESCRIPTION = "foo description"


class PlatformMetaBase(DeclarativeMeta):
    pass


class PlatformBase(object):
    pass


from Ligare.database.testing.config import inmemory_database_config

PlatformBase = declarative_base(cls=PlatformBase, metaclass=PlatformMetaBase)
FeatureFlag = FeatureFlagTable(PlatformBase)


@pytest.fixture()
def feature_flag_session():
    from injector import Injector

    injector = Injector([
        ConfigModule(inmemory_database_config(), DatabaseConfig),
        ScopedSessionModule(bases=[PlatformBase]),  # pyright: ignore[reportArgumentType]
    ])

    session = injector.get(Session)
    PlatformBase.metadata.create_all(session.bind)  # pyright: ignore[reportUnknownMemberType,reportAttributeAccessIssue]
    return session


def _create_feature_flag(session: Session):
    session.add(
        FeatureFlag(
            name=_FEATURE_FLAG_TEST_NAME, description=_FEATURE_FLAG_TEST_DESCRIPTION
        )
    )
    session.commit()


def test__feature_is_enabled__defaults_to_false(feature_flag_session: Session):
    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(
        FeatureFlag, feature_flag_session, logger
    )
    _create_feature_flag(feature_flag_session)

    # The value is false because it was not explicitly enabled. This is the database default value.
    is_enabled = db_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert is_enabled == False


@pytest.mark.parametrize("default", [True, False])
def test__feature_is_enabled__uses_default_when_flag_does_not_exist(
    default: bool,
    feature_flag_session: Session,
):
    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(
        FeatureFlag, feature_flag_session, logger
    )

    is_enabled = db_feature_flag_router.feature_is_enabled(
        _FEATURE_FLAG_TEST_NAME, default
    )

    assert is_enabled == default


def test__feature_is_enabled__disallows_empty_name(feature_flag_session: Session):
    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(
        FeatureFlag, feature_flag_session, logger
    )

    with pytest.raises(ValueError):
        _ = db_feature_flag_router.feature_is_enabled("")


@pytest.mark.parametrize("name", [0, False, True, {}, [], (0,)])
def test__feature_is_enabled__disallows_non_string_names(
    name: Any, feature_flag_session: Session
):
    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(
        FeatureFlag, feature_flag_session, logger
    )

    with pytest.raises(TypeError):
        _ = db_feature_flag_router.feature_is_enabled(name)


def test__set_feature_is_enabled__fails_when_flag_does_not_exist(
    feature_flag_session: Session,
):
    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(
        FeatureFlag, feature_flag_session, logger
    )

    with pytest.raises(LookupError):
        db_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, True)


def test__set_feature_is_enabled__disallows_empty_name(feature_flag_session: Session):
    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(
        FeatureFlag, feature_flag_session, logger
    )
    _create_feature_flag(feature_flag_session)

    with pytest.raises(ValueError):
        db_feature_flag_router.set_feature_is_enabled("", False)


@pytest.mark.parametrize("name", [0, False, True, {}, [], (0,)])
def test__set_feature_is_enabled__disallows_non_string_names(
    name: Any, feature_flag_session: Session
):
    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(
        FeatureFlag, feature_flag_session, logger
    )

    with pytest.raises(TypeError):
        db_feature_flag_router.set_feature_is_enabled(name, False)


@pytest.mark.parametrize("enable", [True, False])
def test__set_feature_is_enabled__sets_correct_value(
    enable: bool, feature_flag_session: Session
):
    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(
        FeatureFlag, feature_flag_session, logger
    )
    _create_feature_flag(feature_flag_session)

    db_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, enable)
    is_enabled = db_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert is_enabled == enable


@pytest.mark.parametrize("enable", [True, False])
def test__set_feature_is_enabled__caches_flags(enable: bool, mocker: MockerFixture):
    session_mock = mocker.patch("sqlalchemy.orm.session.Session")
    session_query_mock = mocker.patch("sqlalchemy.orm.session.Session.query")
    session_mock.query = session_query_mock

    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(FeatureFlag, session_mock, logger)

    db_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, enable)
    _ = db_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)
    _ = db_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert session_query_mock.call_count == 1


@pytest.mark.parametrize("check_cache", [(True, 1), (False, 0)])
def test__feature_is_enabled__checks_cache(
    check_cache: Tuple[bool, int], mocker: MockerFixture
):
    session_mock = mocker.patch("sqlalchemy.orm.session.Session")
    feature_is_enabled_mock = mocker.patch(
        "Ligare.platform.feature_flag.caching_feature_flag_router.CachingFeatureFlagRouter.feature_is_cached"
    )
    _ = mocker.patch(
        "Ligare.platform.feature_flag.caching_feature_flag_router.CachingFeatureFlagRouter.set_feature_is_enabled"
    )

    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(FeatureFlag, session_mock, logger)

    _ = db_feature_flag_router.feature_is_enabled(
        _FEATURE_FLAG_TEST_NAME, False, check_cache[0]
    )

    assert feature_is_enabled_mock.call_count == check_cache[1]


@pytest.mark.parametrize("check_cache", [(True, 0), (False, 1)])
def test__feature_is_enabled__sets_cache(
    check_cache: Tuple[bool, int], mocker: MockerFixture
):
    session_mock = mocker.patch("sqlalchemy.orm.session.Session")
    feature_is_enabled_mock = mocker.patch(
        "Ligare.platform.feature_flag.caching_feature_flag_router.CachingFeatureFlagRouter.feature_is_cached"
    )
    set_feature_is_enabled_mock = mocker.patch(
        "Ligare.platform.feature_flag.caching_feature_flag_router.CachingFeatureFlagRouter.set_feature_is_enabled"
    )
    feature_is_enabled_mock.return_value = True

    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(FeatureFlag, session_mock, logger)

    _ = db_feature_flag_router.feature_is_enabled(
        _FEATURE_FLAG_TEST_NAME, False, check_cache[0]
    )

    assert set_feature_is_enabled_mock.call_count == check_cache[1]


@pytest.mark.parametrize("enable", [True, False])
def test__set_feature_is_enabled__resets_cache_when_flag_enable_is_set(
    enable: bool, mocker: MockerFixture
):
    session_mock = mocker.patch("sqlalchemy.orm.session.Session")
    session_query_mock = mocker.patch("sqlalchemy.orm.session.Session.query")
    session_mock.query = session_query_mock

    logger = logging.getLogger("FeatureFlagLogger")
    db_feature_flag_router = DBFeatureFlagRouter(FeatureFlag, session_mock, logger)

    db_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, enable)
    _ = db_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)
    first_value = db_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    db_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, not enable)
    _ = db_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)
    second_value = db_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert session_query_mock.call_count == 2
    assert first_value == enable
    assert second_value == (not enable)



import logging
from typing import Any

import pytest
from Ligare.platform.feature_flag.caching_feature_flag_router import (
    CachingFeatureFlagRouter,
)
from mock import MagicMock
from pytest import LogCaptureFixture

_FEATURE_FLAG_TEST_NAME = "foo_feature"


def test__feature_is_enabled__disallows_empty_name():
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    with pytest.raises(ValueError):
        _ = caching_feature_flag_router.feature_is_enabled("")


@pytest.mark.parametrize("name", [0, False, True, {}, [], (0,)])
def test__feature_is_enabled__disallows_non_string_names(name: Any):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    with pytest.raises(TypeError):
        _ = caching_feature_flag_router.feature_is_enabled(name)


def test__set_feature_is_enabled__disallows_empty_name():
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    with pytest.raises(ValueError):
        caching_feature_flag_router.set_feature_is_enabled("", False)


@pytest.mark.parametrize("name", [0, False, True, {}, [], (0,)])
def test__set_feature_is_enabled__disallows_non_string_names(name: Any):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    with pytest.raises(TypeError):
        caching_feature_flag_router.set_feature_is_enabled(name, False)


@pytest.mark.parametrize("value", [None, "", "False", "True", 0, 1, -1, {}, [], (0,)])
def test__set_feature_is_enabled__disallows_non_bool_values(value: Any):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    with pytest.raises(TypeError) as e:
        _ = caching_feature_flag_router.set_feature_is_enabled(
            _FEATURE_FLAG_TEST_NAME, value
        )

    assert e.match("`is_enabled` must be a boolean")


@pytest.mark.parametrize("value", [True, False])
def test__set_feature_is_enabled__sets_correct_value(value: bool):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    caching_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, value)
    is_enabled = caching_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert is_enabled == value


def test__feature_is_enabled__defaults_to_false_when_flag_does_not_exist():
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    is_enabled = caching_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert is_enabled == False


@pytest.mark.parametrize("value", [True, False])
def test__set_feature_is_enabled__caches_new_flag(value: bool):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    caching_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, value)
    assert _FEATURE_FLAG_TEST_NAME in caching_feature_flag_router._feature_flags  # pyright: ignore[reportPrivateUsage]
    is_enabled = caching_feature_flag_router._feature_flags.get(_FEATURE_FLAG_TEST_NAME)  # pyright: ignore[reportPrivateUsage]
    assert is_enabled == value


@pytest.mark.parametrize("value", [True, False])
def test__feature_is_enabled__uses_cache(value: bool):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    mock_dict = MagicMock()
    caching_feature_flag_router._feature_flags = mock_dict  # pyright: ignore[reportPrivateUsage]
    caching_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, value)

    _ = caching_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)
    _ = caching_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert mock_dict.get.call_count == 3


@pytest.mark.parametrize("enable", [True, False])
def test__set_feature_is_enabled__resets_cache_when_flag_enable_is_set(enable: bool):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    caching_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, enable)
    _ = caching_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)
    first_value = caching_feature_flag_router.feature_is_enabled(
        _FEATURE_FLAG_TEST_NAME
    )

    caching_feature_flag_router.set_feature_is_enabled(
        _FEATURE_FLAG_TEST_NAME, not enable
    )
    _ = caching_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)
    second_value = caching_feature_flag_router.feature_is_enabled(
        _FEATURE_FLAG_TEST_NAME
    )

    assert first_value == enable
    assert second_value == (not enable)


@pytest.mark.parametrize("value", [True, False])
def test__set_feature_is_enabled__notifies_when_setting_new_flag(
    value: bool,
    caplog: LogCaptureFixture,
):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    caching_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, value)

    assert f"Setting new feature flag '{_FEATURE_FLAG_TEST_NAME}' to `{value}`." in {
        record.msg for record in caplog.records
    }


@pytest.mark.parametrize(
    "first_value,second_value,expected_log_msg",
    [
        [
            True,
            True,
            f"Tried to change feature flag value for '{_FEATURE_FLAG_TEST_NAME}' to the same value. It is already enabled.",
        ],
        [
            False,
            False,
            f"Tried to change feature flag value for '{_FEATURE_FLAG_TEST_NAME}' to the same value. It is already disabled.",
        ],
        [
            True,
            False,
            f"Changing feature flag value for '{_FEATURE_FLAG_TEST_NAME}' from `True` to `False`.",
        ],
        [
            False,
            True,
            f"Changing feature flag value for '{_FEATURE_FLAG_TEST_NAME}' from `False` to `True`.",
        ],
    ],
)
def test__set_feature_is_enabled__notifies_when_changing_flag(
    first_value: bool,
    second_value: bool,
    expected_log_msg: str,
    caplog: LogCaptureFixture,
):
    logger = logging.getLogger("FeatureFlagLogger")
    caching_feature_flag_router = CachingFeatureFlagRouter(logger)

    caching_feature_flag_router.set_feature_is_enabled(
        _FEATURE_FLAG_TEST_NAME, first_value
    )
    caching_feature_flag_router.set_feature_is_enabled(
        _FEATURE_FLAG_TEST_NAME, second_value
    )

    assert expected_log_msg in {record.msg for record in caplog.records}



from dataclasses import dataclass
from enum import Enum
from logging import Logger
from typing import Generic, Protocol, Sequence, Type, TypeVar, cast

from injector import inject
from sqlalchemy.orm import class_mapper  # pyright: ignore[reportUnknownVariableType]
from sqlalchemy.orm import ColumnProperty, RelationshipProperty
from sqlalchemy.orm.mapper import Mapper
from sqlalchemy.orm.scoping import ScopedSession
from typing_extensions import override

from . import Role as DbRole
from . import User as DbUser


@dataclass
class UserId:
    user_id: int
    username: str


class Role(Enum):
    @override
    def __str__(self):
        return self.name


TRole = TypeVar("TRole", bound=Role, covariant=True)


class UserMixin(Protocol[TRole]):
    id: UserId

    # property resolves invariance error by telling
    # pyright this is an immutable "field"
    @property
    def roles(self) -> Sequence[TRole]: ...

    def __init__(self, id: UserId, roles: Sequence[TRole] | None = None) -> None:
        super().__init__()


TUserMixin = TypeVar("TUserMixin", bound=UserMixin[Role])


class UserLoader(Generic[TUserMixin]):
    """
    Class intended for user with FlaskLogin. FlaskLogin is not required.
    """

    @inject
    def __init__(
        self,
        loader: type[TUserMixin],
        roles: Type[Enum],
        user_table: Type[DbUser],
        role_table: Type[DbRole],
        scoped_session: ScopedSession,
        log: Logger,
    ) -> None:
        """
        Load and optionally create a user.

        :param Callable[[UserId, list[Enum]], None] loader: A callback that receives the create user information including its roles.
        :param Type[Enum] roles: An enum representing possible roles for a user.
        :param Type[DbUser] user_table: The SQLAlchemy table type for a user.
        :param Type[DbRole] role_table: The SQLAlchemy table type for a role.
        :param ScopedSession scoped_session: The SQLAlchemy connection scope.
        :param Logger log: A Logger instance.
        """
        self._loader = loader
        self._roles = roles
        self._user_table = user_table
        self._role_table = role_table
        self._scoped_session = scoped_session
        self._log = log
        super().__init__()

    def user_loader(
        self, username: str, default_role: Enum | None, create_if_new_user: bool = False
    ) -> None | TUserMixin:
        """
        Load a user and its roles from the database.

        :param str username: The users username.
        :param Enum default_role: The default role assigned to the user when `create_if_new_user` is true and the user does not already exist.
        :param bool create_if_new_user: Create the user in the database if it does not already exist, defaults to False.

        :raises AssertionError: If a user is loaded from the database, but it is not of the type specified by user_table.
        """
        self._log.debug(f'Loading user "{username}"')

        if not username:
            self._log.warning("`username` is empty. Skipping load.")
            return

        with self._scoped_session() as session:
            try:
                # SQLAlchemy generates invalid SQL when attempting an implicit
                # LEFT JOIN between user -> user_role and user_role -> role.
                # As such, we handle the join explicitly by extracting the property
                # relationships and referencing the relevant columns.
                user_table_property_mapper = cast(
                    Mapper, class_mapper(self._user_table)
                )
                user_table_properties = cast(
                    "list[RelationshipProperty[DbRole] | ColumnProperty]",
                    user_table_property_mapper.iterate_properties,  # pyright: ignore[reportUnknownMemberType]
                )
                # Only extract the secondary join table (user_role).
                # We find this by matching on the `role` table relationship
                # from the `user` table.
                user_role_table = next(
                    relationship.secondary
                    for relationship in user_table_properties
                    if (entity := getattr(relationship, "entity", None))
                    and entity.class_ == self._role_table
                )

                user = (
                    session.query(self._user_table)
                    # These two `outerjoin` calls define the explicit
                    # join conditions between the `user` and `role` tables,
                    # and the secondary join table `user_role`.
                    .outerjoin(
                        user_role_table,
                        user_role_table.c.user_id == self._user_table.user_id,
                    )
                    .outerjoin(
                        self._role_table,
                        user_role_table.c.role_id == self._role_table.role_id,
                    )
                    .filter(self._user_table.username == username)
                    .one_or_none()
                )

                self._log.debug(f'Queried for "{username}" in database')

                if create_if_new_user and user is None:
                    self._log.info(
                        f'User "{username}" does not already exist in the database'
                    )

                    if default_role is None:
                        user = self._user_table(username=username)
                    else:
                        role = (
                            session.query(self._role_table)
                            .filter(self._role_table.role_name == default_role.name)
                            .one()
                        )

                        user = self._user_table(username=username, roles=[role])

                    session.add(user)
                    session.commit()

                    self._log.info(f'User "{username}" added to the database')
                elif user is None:
                    self._log.info(
                        "Username does not exist. Refusing to create a new user. This happens when a username (likely from a cookie)\
        cannot be found in the database, and the process should not create a new user."
                    )
                    return
                else:
                    self._log.debug(f'User "{username}" loaded from the database')

                if not isinstance(user, self._user_table):
                    raise AssertionError(
                        f"User object queried from database is unexpected type `{type(user)}`. Expected type `{type(self._user_table)}`."
                    )

                user_roles = dict(self._roles.__members__.items())
                roles = cast(
                    Sequence[Role],
                    [user_roles[user_role.role_name] for user_role in user.roles],
                )

                return self._loader(
                    UserId(
                        user_id=user.user_id,
                        username=user.username,
                    ),
                    roles,
                )
            except:
                self._log.exception(
                    f'Error when loading user "{username}" from the database.'
                )
                raise



import ast
import inspect

from Ligare.platform.feature_flag.feature_flag_router import FeatureFlagRouter
from typing_extensions import override

_FEATURE_FLAG_TEST_NAME = "foo_feature"


class TestFeatureFlagRouter(FeatureFlagRouter):
    @override
    def set_feature_is_enabled(self, name: str, is_enabled: bool) -> None:
        return super().set_feature_is_enabled(name, is_enabled)

    @override
    def feature_is_enabled(self, name: str, default: bool = False) -> bool:
        return super().feature_is_enabled(name, default)


class NotifyingFeatureFlagRouter(TestFeatureFlagRouter):
    def __init__(self) -> None:
        self.notification_count = 0
        super().__init__()

    @override
    def _notify_change(
        self, name: str, new_value: bool, old_value: bool | None
    ) -> None:
        self.notification_count += 1
        return super()._notify_change(name, new_value, old_value)


def test___notify_change__is_a_noop():
    test_feature_flag_router = TestFeatureFlagRouter()
    source = inspect.getsource(test_feature_flag_router._notify_change)  # pyright: ignore[reportPrivateUsage]
    # strip the leading \t because the method is in a class
    # and the AST parse will fail because of unexpected indent
    parsed = ast.parse(source.strip())
    function_node = parsed.body[0]
    # [0] is the docblock
    function_body_nodes = function_node.body[1:]  # pyright: ignore[reportUnknownMemberType,reportUnknownVariableType,reportAttributeAccessIssue]

    # empty means the function doesn't contain any code
    assert not function_body_nodes


def test__set_feature_is_enabled__does_not_notify():
    notifying_feature_flag_router = NotifyingFeatureFlagRouter()

    notifying_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, True)

    assert notifying_feature_flag_router.notification_count == 0


def test__feature_is_enabled__does_not_notify():
    notifying_feature_flag_router = NotifyingFeatureFlagRouter()

    _ = notifying_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert notifying_feature_flag_router.notification_count == 0


def test__feature_is_enabled__does_not_notify_after_flag_is_set():
    notifying_feature_flag_router = NotifyingFeatureFlagRouter()

    notifying_feature_flag_router.set_feature_is_enabled(_FEATURE_FLAG_TEST_NAME, True)
    notifying_feature_flag_router.notification_count = 0

    _ = notifying_feature_flag_router.feature_is_enabled(_FEATURE_FLAG_TEST_NAME)

    assert notifying_feature_flag_router.notification_count == 0



from abc import ABC
from typing import List, Type, cast

from Ligare.database.schema.metabase import get_schema_from_metabase
from sqlalchemy import Column, ForeignKey, Integer, Unicode
from sqlalchemy.ext.declarative import DeclarativeMeta
from sqlalchemy.orm import RelationshipProperty, relationship
from typing_extensions import override


class Role(ABC):
    __tablename__: str

    def __init__(  # pyright: ignore[reportMissingSuperCall]
        self,
        /,
        role_id: int = 0,
        role_name: str = "",
        users: "list[User]" = [],  # pyright: ignore[reportCallInDefaultInitializer]
    ) -> None:
        raise NotImplementedError(
            f"`{Role.__class__.__name__}` should only be used for type checking."
        )

    role_id: int
    role_name: str
    users: "list[User]"


class User(ABC):
    def __init__(  # pyright: ignore[reportMissingSuperCall]
        self,
        /,
        user_id: int = 0,
        username: str = "",
        roles: "list[Role]" = [],  # pyright: ignore[reportCallInDefaultInitializer]
    ) -> None:
        raise NotImplementedError(
            f"`{User.__class__.__name__}` should only be used for type checking."
        )

    __tablename__: str
    user_id: int
    username: str
    roles: "list[Role]"


class UserRole(ABC):
    def __init__(  # pyright: ignore[reportMissingSuperCall]
        self, /, user_id: int = 0, role_id: int = 0
    ) -> None:
        raise NotImplementedError(
            f"`{UserRole.__class__.__name__}` should only be used for type checking."
        )

    user_id: int
    role_id: int


def get_table_str(tablename: str, base: Type[DeclarativeMeta]):
    schema = get_schema_from_metabase(base)
    schema_str = f"{schema}." if schema else ""
    return f"{schema_str}{tablename}"


class RoleTable:
    def __new__(cls, base: Type[DeclarativeMeta]) -> type[Role]:
        class _Role(base):
            """
            Roles that may be used for access validation.
            """

            __tablename__ = "role"

            role_id = Column("role_id", Integer, primary_key=True)
            role_name = Column("role_name", Unicode, nullable=False, unique=True)

            users: "RelationshipProperty[DeclarativeMeta]" = relationship(
                "_User",
                secondary=get_table_str("user_role", base),
                back_populates="roles",
            )

            @override
            def __repr__(self) -> str:
                return "<Role (role_id=%d, role_name=%s)>" % (
                    self.role_id,
                    self.role_name,
                )

        return cast(type[Role], _Role)


class UserTable:
    def __new__(cls, base: Type[DeclarativeMeta]) -> type[User]:
        class _User(base):
            __tablename__ = "user"

            user_id = Column("user_id", Integer, primary_key=True)
            username = Column("username", Unicode, nullable=False, unique=True)

            roles: "List[Role] | RelationshipProperty[List[Role]]" = relationship(
                "_Role",
                secondary=get_table_str("user_role", base),
                back_populates="users",
            )

            @override
            def __repr__(self) -> str:
                return "<Role (user_id=%d, username=%s)>" % (
                    self.user_id,
                    self.username,
                )

        return cast(type[User], _User)


class UserRoleTable:
    def __new__(cls, base: Type[DeclarativeMeta]) -> type[UserRole]:
        class _UserRole(base):
            __tablename__ = "user_role"

            user_id = Column(
                "user_id",
                Integer,
                ForeignKey(get_table_str("user.user_id", base)),
                primary_key=True,
            )
            role_id = Column(
                "role_id",
                Integer,
                ForeignKey(get_table_str("role.role_id", base)),
                primary_key=True,
            )

            @override
            def __repr__(self) -> str:
                return "<UserRole (user_id=%d, role_id=%d)>" % (
                    self.user_id,
                    self.role_id,
                )

        return cast(type[UserRole], _UserRole)
